/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2011 by Dolby Laboratories. All rights reserved.

    Description: Dialogue Intelligence test application

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "di.h"

/* Input frame size, in samples */
#define INPUT_FRAME_SIZE            19200

/* Error codes */
#define ERR_INVALID_BIT_DEPTH       -1
#define ERR_CANNOT_OPEN_PCM_INPUT   -2
#define ERR_CANNOT_OPEN_OUTPUT      -3
#define ERR_CANNOT_OPEN_REFERENCE   -4
#define ERR_CANNOT_ALLOCATE_MEMORY  -5
#define ERR_INITIALIZATION_FAILED   -6

/* Conformance test defines */
#define MIN_PASS_RATE               97.0f
#define TEST_PASS                   0
#define TEST_FAIL                   1

/* Private prototypes */
static int conformance_test(char **argv, unsigned int out_file_idx, unsigned int ref_file_idx);
static int parse_args(int argc, char **argv, unsigned int *bit_depth, unsigned int *sample_rate,
                      unsigned int *ref_file_idx, unsigned int *inp_file_idx, unsigned int *out_file_idx);
static void show_help(void);

/* Note on compiler warnings:
 *
 * This test application calls library functions (e.g. fopen()) that can trigger compiler
 * warnings with Microsoft compilers.  Microsoft have special versions of these functions,
 * (e.g. fopen_s()) but these versions are not used as they aren't cross-platform.
 *
 * To disable these compiler warnings on Microsoft compilers, define the
 * preprocessor symbol  _CRT_SECURE_NO_WARNINGS.
 */

/* Public functions */
/*******************************************************************************************
int main(int argc, char **argv)
-Purpose:
    Main entry point for the Dialogue Intelligence test application.
-Parameters:
    int argc        - (in) Argument count
    char **argv     - (in) Argument vector
-Returns:
    Returns 0 on success, or a negative error code on error.
********************************************************************************************/
int main(int argc, char **argv)
{
    unsigned int    sample_rate;            /* Command line arg: sample rate */
    unsigned int    bit_depth;              /* Command line arg: bit depth */
    unsigned int    ref_file_idx;           /* Command line arg: reference file (index into argv) */
    unsigned int    inp_file_idx;           /* Command line arg: input file (index into argv) */
    unsigned int    out_file_idx;           /* Command line arg: output file (index into argv) */
    FILE            *p_input, *p_output;    /* File handles for input and output */
    void            *p_di;                  /* Handle to instance of DI */
    unsigned char   p_fixed_pt_samples[INPUT_FRAME_SIZE * 3];   /* Buffer for holding fixed point samples */
    float           p_samples[INPUT_FRAME_SIZE];                /* Buffer for holding floating point samples */
    int             output;                 /* Dialogue Intelligence output */
    char            output8;                /* 8-bit DI output */
    int             is_eof;                 /* Flag indicating end of file */
    size_t          word_size;              /* Input word size, in bytes */
    float           max_sample;             /* Maximum possible input sample value */
    unsigned char   *p;                     /* Temp pointer */
    unsigned int    count;                  /* Frame counter, for UI */
    unsigned int    samples_read;           /* How many input samples were read? */
    int             value;                  /* Temp holder for fixed-float conversion */
    size_t          num_bytes;              /* Number of bytes */
    unsigned int    buffered_samples;       /* Number of samples buffered in DI */
    int             i;                      /* Loop counter */ 

    /* Initialization */
    count = 0;
    is_eof = 0;
    p_input = NULL;
    p_output = NULL;
    buffered_samples = 0;
    bit_depth = 24;
    sample_rate = 48000;
    ref_file_idx = 0;
    inp_file_idx = 0;
    out_file_idx = 0;
        
    /* Parse the command line arguments */
    if (parse_args(argc, argv, &bit_depth, &sample_rate, &ref_file_idx, &inp_file_idx, &out_file_idx) < 0)
    {
        /* Command line error */
        return -1;
    }
    
    /* Determine the word size and maximum sample value */
    switch (bit_depth)
    {
    case 16:    word_size = 2;
                max_sample = 32768.0;
                break;
    case 20:
    case 24:    word_size = 3;
                max_sample = 8388608.0;
                break;
    default:    /* Command line error - should never get here because of parse_args() */
                fprintf(stderr, "ERROR: Invalid bit depth: %u\n", bit_depth);
                return ERR_INVALID_BIT_DEPTH;
    }

    /* Open input pcm file */
    if ((p_input = fopen(argv[inp_file_idx], "rb")) == NULL)
    {
        fprintf(stderr, "ERROR: Cannot open input PCM file: %s\n", argv[inp_file_idx]);
        return ERR_CANNOT_OPEN_PCM_INPUT;
    }
    
    /* Open output file */
    if ((p_output = fopen(argv[out_file_idx], "wb")) == NULL)
    {
        fprintf(stderr, "ERROR: Cannot open output file: %s\n", argv[out_file_idx]);
        (void) fclose(p_input);
        return ERR_CANNOT_OPEN_OUTPUT;
    }
    
    /* Allocate memory for Dialogue Intelligence */
    num_bytes = di_query_mem_size(sample_rate, INPUT_FRAME_SIZE);
    p_di = (void *) malloc(num_bytes);
    if (p_di == NULL)
    {
        fprintf(stderr, "ERROR: Memory allocation failed.\n");
        (void) fclose(p_input);
        (void) fclose(p_output);
        return ERR_CANNOT_ALLOCATE_MEMORY;
    }
    
    /* Configure Dialogue Intelligence */
    if (di_init(p_di, num_bytes, sample_rate, INPUT_FRAME_SIZE) < 0)
    {
        fprintf(stderr, "ERROR: Dialogue Intelligence initialization failed.\n");
        free(p_di);
        (void) fclose(p_input);
        (void) fclose(p_output);
        return ERR_INITIALIZATION_FAILED;
    }

    /* Loop until the end of file.  Due to the latency of DI, the first samples
     * produce no output (di_process() == INVALID).  Hence the equivalent number
     * of zeros must be inserted into DI in order to flush out the final outputs.
     */
    while ((is_eof == 0) || (buffered_samples > 0))
    {    
        /* Update the display counter */
        count++;
        if ((count % 10) == 0)
        {
            fprintf(stdout, "Frame #%u\n", count);
        }
        
        /* Check if samples are sourced from the file, or if we're feeding zeros
         * at the end of the file to flush out DI.
         */
        if (is_eof == 0)
        {    
            /* Read samples from the file */
            samples_read = (unsigned int) fread(p_fixed_pt_samples, word_size, INPUT_FRAME_SIZE, p_input);
            
            /* Check if we are at the end of the file */
            if (samples_read != INPUT_FRAME_SIZE)
            {
                is_eof = 1;
                
                /* The final frame could be a partial frame.  It is processed by di_process() in the same manner as a full frame. */
            }
        }
        else
        {
            /* Insert zeros */
            samples_read = (buffered_samples < INPUT_FRAME_SIZE) ? buffered_samples : INPUT_FRAME_SIZE;
            buffered_samples -= samples_read;
            memset(p_samples, 0, samples_read * sizeof(float));
        }
        
        /* Check that we have samples to process */
        if (samples_read > 0)
        {   
            /* Convert the samples to float */
            p = p_fixed_pt_samples;
            switch (bit_depth)
            {
            case 16:    /* Convert 16-bit inputs */
                        for (i = 0; i < (int) samples_read; i++)
                        {
                            value = (int) ( (p[1] << 8) |
                                             p[0]
                                          );   
				            if ((value & 0x00008000) != 0)
				            {
				                value |= 0xffff0000;
				            }
				            p_samples[i] = (float)value / max_sample;
                            p += word_size;
                        }
                        break;
                        
            case 20:    /* Convert 20-bit inputs */
                        for (i = 0; i < (int) samples_read; i++)
                        {
				            value = (int) ( (p[2] << 16) |
				                            (p[1] << 8)  |
				                            (p[0] & 0xf0)
				                          );
				            if ((value & 0x00800000) != 0)
				            {
				                value |= 0xff000000;
				            }
				            p_samples[i] = (float)value / max_sample;
                            p += word_size;
                        }
                        break;
                        
            case 24:    /* Convert 24-bit inputs */
                        for (i = 0; i < (int) samples_read; i++)
                        {
				            value = (int) ( (p[2] << 16) |
				                            (p[1] << 8)  |
				                             p[0]
				                          ); 
				            if ((value & 0x00800000) != 0)
				            {
				                value |= 0xff000000;
				            }
				            p_samples[i] = (float)value / max_sample;
                            p += word_size;
                        }
                        break;

            default:    /* Should never get here */
                        free(p_di);
                        (void) fclose(p_input);
                        (void) fclose(p_output);
                        return ERR_INVALID_BIT_DEPTH;
            }
            
            /* Feed samples to DI */
            output = di_process(p_di, p_samples, samples_read);
            
            /* Check for invalid outputs (i.e. DI is buffering audio) */
            if (output == INVALID)
            {
                buffered_samples += samples_read;
            }
            else
            {
                /* Save the output */
                output8 = (char) ((output == SPEECH) ? 0x01 : 0x00);
                for (i = 0; i < (int) samples_read; i++)
                {
                    (void) fwrite(&output8, 1, 1, p_output);
                }
            }
        } /* if (samples_read > 0) */
    } /* while ((is_eof == 0) || (buffered_samples > 0)) */

    /* Close the input and output files */
    (void) fclose(p_input);
    (void) fclose(p_output);
    
    /* Free the memory allocated for DI */
    free(p_di);

    /* Run the conformance test, if requested */
    if (ref_file_idx != 0)
    {
        fprintf(stdout, "Conformance test...\n");
        
        if (conformance_test(argv, out_file_idx, ref_file_idx) == TEST_PASS)
        {
            fprintf(stdout, "*** PASS ***\n");
        }
        else
        {
            fprintf(stdout, "*** FAIL ***\n");
        }
    }
    else
    {
        fprintf(stdout, "Conformance test not run.\n");
    }
    return 0;
}

/* Private functions */
/*******************************************************************************************
static int conformance_test(char **argv, unsigned int out_file_idx, unsigned int ref_file_idx)
-Purpose:
    Run the conformance test using file names specified on the command line.  Refer to the
    accompanying code user's guide for a full description of the conformance test.
-Parameters:
    char **argv                 - (in) Argument vector
    unsigned int out_file_idx   - (in) Index to the output file on the command line arguments vector
    unsigned int out_file_idx   - (in) Index to the output file on the command line arguments vector
-Returns:
    TEST_PASS, TEST_FAIL or a negative error code
********************************************************************************************/
static int conformance_test(char **argv, unsigned int out_file_idx, unsigned int ref_file_idx)
{
    FILE            *p_output;              /* File handle for output */
    FILE            *p_reference;           /* File handle for the reference */
    int             is_eof;                 /* Flag indicating end of file */
    size_t          num_output_read;        /* Number of output values read */
    size_t          num_reference_read;     /* Number of reference values read */
    unsigned char   reference;              /* Reference value */
    unsigned char   output;                 /* DI output value */
    int             passes;                 /* Number of true results */
    int             total_results;          /* Number of results */
    float           pass_rate;              /* Pass rate (%) */
    int             result;                 /* Test result */
    
    /* Initialization */
    is_eof = 0;
    passes = 0;
    total_results = 0;
    result = -1;
    num_output_read = 0;
    num_reference_read = 0;
    
    /* Open the output file for reading */
    if ((p_output = fopen(argv[out_file_idx], "rb")) == NULL)
    {
        fprintf(stderr, "Cannot open output file: %s\n", argv[out_file_idx]);
        return ERR_CANNOT_OPEN_OUTPUT;
    }

    /* Open the reference file */
    if ((p_reference = fopen(argv[ref_file_idx], "rb")) == NULL)
    {
        fprintf(stderr, "Cannot open reference file: %s\n", argv[ref_file_idx]);
        (void) fclose(p_output);
        return ERR_CANNOT_OPEN_REFERENCE;
    }

    /* Loop through every 8-bit value in the file */
    while (is_eof == 0)
    {
        /* Read the expected output from the output file */
        num_output_read = fread(&output, sizeof(unsigned char), 1, p_output);
        
        /* Read the expected output from the reference */
        num_reference_read = fread(&reference, sizeof(unsigned char), 1, p_reference);
        
        /* Check that fread succeeded */
        if ((num_output_read == 1) && (num_reference_read == 1))
        {
            if (output == reference)
            {
                passes++;
            }
            total_results++;
        }
        else
        {
            /* End of file */
            is_eof = 1;
        }
    }
    
    /* Perform the final calculations */
    pass_rate = 100.0f * ((float) passes) / ((float) total_results);
    if ((num_output_read != 0) || (num_reference_read != 0))
    {
        fprintf(stdout, "   Output and Reference are different sizes!\n");
        result = TEST_FAIL;
    }
    else
    {
        fprintf(stdout, "   Pass rate : %3.2f%%\n", pass_rate);
        if (pass_rate >= MIN_PASS_RATE)
        {
            result = TEST_PASS;
        }
        else
        {
            result = TEST_FAIL;
        }
    }   
    
    /* Close the files */
    (void) fclose(p_output);
    (void) fclose(p_reference);
    
    return result;
}

/*******************************************************************************************
static int parse_args(int argc, char **argv, unsigned int *bit_depth, unsigned int *sample_rate,
                      unsigned int *ref_file_idx, unsigned int *inp_file_idx, unsigned int *out_file_idx)
-Purpose:
    Parse the command line arguments.  Will display help if command line errors are detected.
-Parameters:
    int argc                    - (in) Argument count
    char **argv                 - (in) Argument vector
    unsigned int *bit_depth     - (out) Bit depth (default 24)
    unsigned int *sample_rate   - (out) Sampling rate, in Hz (default 48000)
    unsigned int *ref_file_idx  - (out) Reference file index (0 if not specified)
    unsigned int *inp_file_idx  - (out) Input PCM file index
    unsigned int *out_file_idx  - (out) Output file index
-Returns:
    0 if successful, -1 on command line error.
********************************************************************************************/
static int parse_args(int argc, char **argv, unsigned int *bit_depth, unsigned int *sample_rate,
                      unsigned int *ref_file_idx, unsigned int *inp_file_idx, unsigned int *out_file_idx)
{
    unsigned int i;
    
    /* Set the defaults */
    *bit_depth      = 24;
    *sample_rate    = 48000;
    *ref_file_idx   = 0;
    *inp_file_idx   = 0;
    *out_file_idx   = 0;
    
    /* Check argc is valid */
    if (argc < 0)
    {
        fprintf(stderr, "ERROR: No arguments specified.\n");
        show_help();
        return -1;
    }
    
    /* Loop through every argument */
    for (i = 1; i < (unsigned) argc; i++)
    {
        /* Ensure the length is correct.
         * We're expecting a "-b", "-f" etc.
         */
        if (strlen(argv[i]) != 2)
        {
            fprintf(stderr, "ERROR: Invalid format: \"%s\".\n", argv[i]);
            show_help();
            return -1;
        }
        
        /* Ensure the first character is a dash */
        if (argv[i][0] != '-')
        {
            fprintf(stderr, "ERROR: Invalid format: \"%s\".\n", argv[i]);
            show_help();
            return -1;
        }
        
        /* Ensure there is another argument following */
        if ((unsigned) argc < i+1)
        {
            fprintf(stderr, "ERROR: No argument following: \"%s\".\n", argv[i]);
            show_help();
            return -1;
        }
        
        /* Get the option character, and respond accordingly */
        switch (argv[i][1])
        {
        case 'h':   /* Show help */
                    show_help();
                    return -1;
        
        case 'b':   /* Set a bit depth */
                    
                    /* Ensure the next argument is the correct length */
                    if (strlen(argv[i+1]) != 2)
                    {
                        fprintf(stderr, "ERROR: Invalid argument: \"%s %s\".\n", argv[i], argv[i+1]);
                        show_help();
                        return -1;
                    }
                    
                    /* Parse the next argument */
                    if ((argv[i+1][0] == '1') &&
                        (argv[i+1][1] == '6'))
                    {
                        *bit_depth = 16;
                    }
                    else if ((argv[i+1][0] == '2') &&
                             (argv[i+1][1] == '0'))
                    {
                        *bit_depth = 20;
                    }
                    else if ((argv[i+1][0] == '2') &&
                             (argv[i+1][1] == '4'))
                    {
                        *bit_depth = 24;
                    }
                    else
                    {
                        fprintf(stderr, "ERROR: Invalid bit depth: \"%s\".\n", argv[i+1]);
                        show_help();
                        return -1;
                    }
                    
                    break;
                    
        case 'f':   /* Set a sampling rate */
                    
                    /* Ensure the next argument is the correct length */
                    if (strlen(argv[i+1]) != 5)
                    {
                        fprintf(stderr, "ERROR: Invalid argument: \"%s %s\".\n", argv[i], argv[i+1]);
                        show_help();
                        return -1;
                    }
                    
                    /* Parse the next argument */
                    if ((argv[i+1][0] == '3') &&
                        (argv[i+1][1] == '2') &&
                        (argv[i+1][2] == '0') &&
                        (argv[i+1][3] == '0') &&
                        (argv[i+1][4] == '0'))
                    {
                        *sample_rate = 32000;
                    }
                    else if ((argv[i+1][0] == '4') &&
                             (argv[i+1][1] == '4') &&
                             (argv[i+1][2] == '1') &&
                             (argv[i+1][3] == '0') &&
                             (argv[i+1][4] == '0'))
                    {
                        *sample_rate = 44100;
                    }
                    else if ((argv[i+1][0] == '4') &&
                             (argv[i+1][1] == '8') &&
                             (argv[i+1][2] == '0') &&
                             (argv[i+1][3] == '0') &&
                             (argv[i+1][4] == '0'))
                    {
                        *sample_rate = 48000;
                    }
                    else
                    {
                        fprintf(stderr, "ERROR: Invalid sample rate: \"%s\".\n", argv[i+1]);
                        show_help();
                        return -1;
                    }
                    
                    break;
                    
        case 'r':   /* Specify a reference file */
        case 'i':   /* Specify an input file */
        case 'o':   /* Specify an output file */
                    
                    /* Ensure the next argument is at least one character long */
                    if (strlen(argv[i+1]) < 1)
                    {
                        fprintf(stderr, "ERROR: Invalid format: \"%s %s\".\n", argv[i], argv[i+1]);
                        show_help();
                        return -1;
                    }
                    
                    /* Save the file name index */
                    if (argv[i][1] == 'r')
                    {
                        *ref_file_idx = i+1;
                    }
                    else if (argv[i][1] == 'i')
                    {
                        *inp_file_idx = i+1;
                    }
                    else if (argv[i][1] == 'o')
                    {
                        *out_file_idx = i+1;
                    }
                    
                    break;
                    
        default:    /* Unrecognized option */
                    fprintf(stderr, "ERROR: Unrecognized option: \"%s\".\n", argv[i]);
                    show_help();
                    return -1;
                                  
        }
        
        /* The next argument has already been parsed.  Skip over it. */
        i++;
        
    }
    
    /* Check for required options */
    if ((*inp_file_idx == 0) ||
        (*out_file_idx == 0))
    {
        fprintf(stderr, "ERROR: Input and/or output file(s) not specified.\n");
        show_help();
        return -1;
    }
    
    return 0;

}

/*******************************************************************************************
static void show_help(void)
-Purpose:
    Display the help message for the Dialogue Intelligence test application.
-Parameters:
    None
-Returns:
    None
********************************************************************************************/
static void show_help(void)
{

    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "+------------------------------------------------------------------------------+\n");
    fprintf(stdout, "| Dialogue Intelligence Test Application                                       |\n");
    fprintf(stdout, "| Dialogue Intelligence Version %5s                                          |\n", DI_VERSION);   
    fprintf(stdout, "| Copyright 2011 by Dolby Laboratories. All rights reserved.                   |\n");
    fprintf(stdout, "+------------------------------------------------------------------------------+\n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "This application extracts audio from an input PCM file, and passes that         \n");
    fprintf(stdout, "audio through Dialogue Intelligence (DI).  The output from DI,                  \n");
    fprintf(stdout, "a binary SPEECH/OTHER decision variable, is saved in an output file.            \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "This application properly accounts for the latency of DI.  As                   \n");
    fprintf(stdout, "DI does not produce any classification outputs for the first 2.048s of          \n");
    fprintf(stdout, "input, no outputs are written during this period.  Additionally, the application\n");
    fprintf(stdout, "will suffix the PCM audio with a 2.048s silence period that allows the final    \n");
    fprintf(stdout, "classification results to be extracted from DI.                                 \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "This test application is capable of running the DI conformance test specified in\n");
    fprintf(stdout, "the manual.  If a \"reference\" file is specified this test will be run.        \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "The input PCM file is a binary file containing a single channel of PCM.  The    \n");
    fprintf(stdout, "sample format is 16, 20 or 24-bit, and the sampling rate is either 32,          \n");
    fprintf(stdout, "44.1 or 48kHz.  Byte order is little endian.  20-bit data, if used, is stored in\n");
    fprintf(stdout, "the top 20 bits of 24-bit words leaving the bottom 4 bits set to zero.          \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "The output and reference files are binary files each containing an array of     \n");
    fprintf(stdout, "8-bit values, one value per input sample.  The values are either 0x01 (speech)  \n");
    fprintf(stdout, "or 0x00 (other).                                                                \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "Usage: di-test [-h] [-b 16|20|24] [-f 32000|44100|48000] [-r filename]          \n");
    fprintf(stdout, "            -i filename -o filename                                             \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "Where:  -h                                                                      \n");
    fprintf(stdout, "            Show the help message                                               \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "        -b 16|20|24                                                             \n");
    fprintf(stdout, "            Specify the bit resolution of the input PCM file (default: 24)      \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "        -f 32000|44100|48000                                                    \n");
    fprintf(stdout, "            Specify the sampling rate of the input file in Hz (default: 48000)  \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "        -r filename                                                             \n");
    fprintf(stdout, "            Specify the name of a reference file.  If -r is missing from the    \n");
    fprintf(stdout, "            command line args, no conformance test is run.                      \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "        -i filename                                                             \n");
    fprintf(stdout, "            Specify the name of an input PCM file.                              \n");
    fprintf(stdout, "                                                                                \n");
    fprintf(stdout, "        -o filename                                                             \n");
    fprintf(stdout, "            Specify the name of a file to store Dialogue Intelligence output    \n");
    
    return;
}
