/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Average spectral tilt feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <math.h>

#include "ast.h"
#include "fft.h"
#include "error.h"

/* Defines */
#define S1_2                    5.482666666666667e3f
#define MINIMUM_DB              -120.0f                 /* Minimum dB value */
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / AST_BLOCK_SIZE)
#define ABS_MAX_LENGTH          16
#define FFT_SIZE                AST_BLOCK_SIZE

/* Instance structure */
typedef struct
{
    float   p_spectral_tilt[BLOCKS_PER_FRAME];  /* Array of spectral tilts per block */
    int     spectral_tilt_idx;                  /* Index into p_spectral_tilt */
    float   p_absmax[ABS_MAX_LENGTH];           /* Array of block absolute maximums. */
    int     absmax_idx;                         /* Index into p_absmax */
    float   p_real[AST_BLOCK_SIZE];             /* Real sample buffer */
    float   p_imag[AST_BLOCK_SIZE];             /* Imag sample buffer */
} AST;

/* Externs */
extern const float p_ast_hamming[];

/* Public functions */
/*******************************************************************************************
size_t ast_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t ast_query_mem_size(void)
{
    return sizeof(AST);
}

/*******************************************************************************************
int ast_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the average spectral tilt feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No ABS_MAX_LENGTH allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of ABS_MAX_LENGTH pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int ast_init(void *p_inst, size_t num_bytes)
{
    int i;
    AST *p_ast;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == ast_query_mem_size());

    /* Retrieve the instance pointer */
    p_ast = (AST *) p_inst;

    /* Initialize the spectral tilt index */
    p_ast->spectral_tilt_idx = 0;

    /* Initialize the spectral tilts */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_ast->p_spectral_tilt[i] = 0.0f;
    }
    
    /* Initialize the absmax index */
    p_ast->absmax_idx = 0;
    
    /* Initialize the absmax array */
    for (i = 0; i < ABS_MAX_LENGTH; i++)
    {
        p_ast->p_absmax[i] = 32768.0f;
    }

    return 0;
}

/*******************************************************************************************
int ast_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Update the average spectral tilt feature for a block of audio.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    float *p_audio          - (in) Pointer to a block of audio.
    int num_samples         - (in) Number of samples contained in p_audio.  Required to be AST_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int ast_process_block(void *p_inst, float *p_audio, int num_samples)
{
    float   *p_real;
    float   *p_imag;
    int     i;
    float   gradient1;
    float   gradient2;
    float   max;
    float   mean;
    float   sum;
    float   temp;
    AST     *p_ast;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == AST_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_ast = (AST *) p_inst;
    
    /* Retrieve the real and imag pointers */
    p_real = p_ast->p_real;
    p_imag = p_ast->p_imag;
    
    /* Find the sum of the amplitudes */
    sum = 0.0f;
    for (i = 0; i < AST_BLOCK_SIZE; i++)
    {
        sum += (float) (fabs(p_audio[i]));
    }
    
    /* Do not perform a feature calculation on "quiet" blocks. */
    if (sum < 10.0f * AST_BLOCK_SIZE)
    {
        /* Set this feature to zero, and return. */
        p_ast->p_spectral_tilt[p_ast->spectral_tilt_idx] = 0.0f;
        p_ast->spectral_tilt_idx++;
        p_ast->spectral_tilt_idx &= (BLOCKS_PER_FRAME - 1);
        return 0;
    }

    /* Find the mean of the block. */
    mean = 0.0f;
    for (i = 0; i < AST_BLOCK_SIZE; i++)
    {
        mean += p_audio[i];
    }
    mean /= (float) AST_BLOCK_SIZE;

    /* Find the absolute maximum of the block. */
    max = 1.0f;
    for (i = 0; i < AST_BLOCK_SIZE; i++)
    {
        temp = (float) fabs(p_audio[i]);
        max = (max > temp) ? max : temp;
    }

    /* Save the absolute maximum */
    p_ast->p_absmax[p_ast->absmax_idx] = max;
    p_ast->absmax_idx++;
    p_ast->absmax_idx &= (ABS_MAX_LENGTH - 1);

    /* Find the absolute maximum of the last ABS_MAX_LENGTH blocks. */
    max = 1.0f;
    for (i = 0; i < ABS_MAX_LENGTH; i++)
    {
        max = (max > p_ast->p_absmax[i]) ? max : p_ast->p_absmax[i];
    }

    /* Remove the mean, apply the hamming window, normalize to the absolute
     * maximum in the last ABS_MAX_LENGTH blocks, and copy into a buffer for the
     * FFT.
     */
    for (i = 0; i < FFT_SIZE; i++)
    {   
        temp = p_audio[i] - mean;
        temp *= p_ast_hamming[i];
        temp /= max;
        
        /* Data is real, so set the imaginary buffer to zero. */
        p_real[i] = temp;
        p_imag[i] = 0.0f;
    }

    /* Calculate FFT */
    (void) fftl(p_real, p_imag, FFT_SIZE);

    /* Calculate ( x .* conj(x) ) */
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        p_real[i] = p_real[i] * p_real[i] + p_imag[i] * p_imag[i];
    }

    /* Calculate the spectral tilt. */
    gradient1 = 0.0f;
    gradient2 = 0.0f;
    sum = 0.0f;
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        if (p_real[i] > 0.0f)
        {
            temp = (float) (10.0 * log10f(p_real[i]));
        }
        else
        {
            temp = MINIMUM_DB;
        }

        gradient1 += temp * (float) i;
        gradient2 += temp;
        sum += (float) i;
    }

    /* Find the gradient */
    gradient1 = (float) (FFT_SIZE/2) * gradient1 - sum * gradient2;
    gradient1 /= ((float) (FFT_SIZE/2) * (float) ((FFT_SIZE/2) - 1) * S1_2);

    /* Save the spectral tilt */
    p_ast->p_spectral_tilt[p_ast->spectral_tilt_idx] = gradient1;
    p_ast->spectral_tilt_idx++;
    p_ast->spectral_tilt_idx &= (BLOCKS_PER_FRAME - 1);

    return 0;
}

/*******************************************************************************************
float ast_process_feature_frame(void *p_inst)
-Purpose:
    Calculate an instance of the average spectral tilt feature for a frame.
-Parameters:
    void *p_inst          - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float ast_process_feature_frame(void *p_inst)
{
    int     i;
    float   return_value;
    float   mean;
    float   skew;
    float   temp;
    AST     *p_ast;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_ast = (AST *) p_inst;

    /* Find the mean of the spectral tilts. */
    mean = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        mean += p_ast->p_spectral_tilt[i];
    }
    mean /= (float) BLOCKS_PER_FRAME;
    
    /* Find the skew of the spectral tilts. */
    skew = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        temp = p_ast->p_spectral_tilt[i] - mean;
        skew += temp * temp * temp;
    }
    skew /= (float) BLOCKS_PER_FRAME;

    return_value = (float) (fabs(skew));
    
    if (return_value > 0.0f)
    {
        return_value = (float) (log(return_value));
    }
    else
    {
        return_value = 0.0f;
    }

    return return_value;

}
