/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Boosting

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdio.h>

#include "bst.h"
#include "error.h"

/* Defines */
#define BOOSTING_ROUNDS         20
#define ACCUMULATE_WINDOW_SIZE  4    /* Length of window for accumulating Dialogue Intelligence results */

/* Instance structure */
typedef struct {
    float    p_prev_boost[ACCUMULATE_WINDOW_SIZE];  /* Array of previous boosting results */
    int      prev_boost_idx;                        /* Index into p_prev_boost */
} BST;

/* Configuration structure */
typedef struct
{
    int     feature;        /* Feature ID */
    float   threshold;      /* Boost threshold */
    float   sign;           /* Boost sign */
    float   weight;         /* Boost weight */
} BOOST_CFG;

/* Boosting configuration */
static const BOOST_CFG p_config[BOOSTING_ROUNDS] =
{
/*    +- Feature
 *    |           +- Threshold
 *    |           |           +- Sign
 *    |           |           |           +- Weight
 *    |           |           |           |           */
    { SFV_IDX,    5.694554f,  1.000000f,  1.123807f },
    { ZCS_IDX,    1.444070f,  1.000000f,  0.765697f },
    { ZCM_IDX,    0.844887f, -1.000000f,  0.412701f },
    { SFV_IDX,    5.807882f,  1.000000f,  0.426570f },
    { LRM_IDX,    0.245440f, -1.000000f,  0.458908f },
    { RPM_IDX,    0.200922f, -1.000000f,  0.326310f },
    { RPM_IDX,    0.196127f, -1.000000f,  0.292287f },
    { SFV_IDX,    5.824250f,  1.000000f,  0.306604f },
    { ZCM_IDX,    0.862523f, -1.000000f,  0.262362f },
    { AST_IDX,  -10.885657f,  1.000000f,  0.147120f },
    { PSC_IDX,    6.042284f,  1.000000f,  0.084314f },
    { SFV_IDX,    5.820881f,  1.000000f,  0.077929f },
    { RPM_IDX,    0.194203f, -1.000000f,  0.046505f },
    { PSC_IDX,    5.974709f,  1.000000f,  0.101153f },
    { SFV_IDX,    5.818805f,  1.000000f,  0.023984f },
    { RPM_IDX,    0.194148f, -1.000000f,  0.014321f },
    { ZCM_IDX,    0.866923f, -1.000000f,  0.045337f },
    { RPM_IDX,    0.194921f, -1.000000f,  0.022334f },
    { SFV_IDX,    5.820068f,  1.000000f,  0.015777f },
    { ZCM_IDX,    0.868066f, -1.000000f,  0.019120f }
};

/* Private prototypes */
static float signf(float value);

/* Public functions */
/*******************************************************************************************
size_t bst_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t bst_query_mem_size(void)
{
    return sizeof(BST);
}

/*******************************************************************************************
int bst_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize an instance of the boosting algorithm.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int bst_init(void *p_inst, size_t num_bytes)
{
    int i;
    BST *p_bst;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == bst_query_mem_size());
    
    /* Retrieve the instance pointer */
    p_bst = (BST *) p_inst;
    
    /* Initialize the array of previous results */
    for (i = 0; i < ACCUMULATE_WINDOW_SIZE; i++)
    {
        p_bst->p_prev_boost[i] = 0.0f;
    }
    p_bst->prev_boost_idx = 0;
    
    return 0;
}

/*******************************************************************************************
int bst_classify(void *p_inst, float *p_features)
-Purpose:
    Cleanup an instance of the boosting algorithm.  It is up to the calling code
    to free the memory allocated to the instance.
-Parameters:
    void *p_inst      - (in/out) Pointer to an instance.
    float *p_features   - (in) Pointer to the feature vector of NUM_FEATURES elements.
-Returns:
    Nothing.
********************************************************************************************/
int bst_classify(void *p_inst, float *p_features)
{
    int     return_value;
    int     boost_round;
    float   sum;
    float   temp;
    int     feature;
    float   threshold;
    float   sign;
    float   weight;
    int     i;
    float   *p_prev_boost;
    BST     *p_bst;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_features != NULL);
    
    /* Retrieve the instance pointer */
    p_bst = (BST *) p_inst;

    /* Retrieve a pointer to the array of previous boost results */
    p_prev_boost = p_bst->p_prev_boost;

    /* Initialize the sum */
    sum = 0.0f;
    
    /* Boosting */
    for (boost_round = 0; boost_round < BOOSTING_ROUNDS; boost_round++)
    {
        /* Get the feature, threshold, weight and sign for this round */
        feature = p_config[boost_round].feature;
        threshold = p_config[boost_round].threshold;
        weight = p_config[boost_round].weight;
        sign = p_config[boost_round].sign;

        /* Perform the boosting */
        temp = p_features[feature];
        temp = sign * signf(temp - threshold);
        temp *= weight;
        sum += temp;
    }

    /* Accumulate current and previous boost results */
    p_prev_boost[p_bst->prev_boost_idx] = sum;
    p_bst->prev_boost_idx++;
    p_bst->prev_boost_idx &= (ACCUMULATE_WINDOW_SIZE-1);
    
    sum = 0.0f;
    for (i = 0; i < ACCUMULATE_WINDOW_SIZE; i++)
    {
        sum += p_prev_boost[i];
    }

    /* Find the overall result */
    sum = signf(sum);
    return_value = ((sum <= 0.0f) ? OTHER : SPEECH);
    return return_value;

}

/* Private functions */
/*******************************************************************************************
static float signf(float value)
-Purpose:
    Find the sign of a floating point value.
-Parameters:
    float value     - (in) Value to find the sign of
-Returns:
    Sign of the input value, either +1.0 or -1.0.
********************************************************************************************/
static float signf(float value)
{
    return ((value < 0.0f) ? -1.0f : 1.0f);
}
