/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Dialogue Intelligence

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <math.h>
#include <string.h>

#include "di.h"
#include "error.h"

/* Defines */
#define INIT_DELAY  ( -3 * MAX_BLOCKS_PER_FRAME )
#define MAX(a,b)    ((a)>(b)?(a):(b))  

/* Public functions */
/*******************************************************************************************
size_t di_query_mem_size(unsigned int sample_rate, unsigned int max_input_samples)
-Purpose:
    Get the size of an instance, in bytes, for a specified sample rate and maximum input
    block size.
-Parameters:
    unsigned int sample_rate         - Sample rate, in Hz (either 32000, 44100 or 48000).
    unsigned int max_input_samples   - Maximum input block size, in samples
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t di_query_mem_size(unsigned int sample_rate, unsigned int max_input_samples)
{
    size_t          sz;
    unsigned int    src_max_input_size = 0;
    unsigned int    delay_line_length;
    
    ERR_CHECKARG((sample_rate == 32000) || (sample_rate == 44100) || (sample_rate == 48000));
    ERR_CHECKARG(max_input_samples > 0);
    
    /* Memory is arranged as:
     * Address 0    DI instance structure
     *              AST instance structure
     *              BST instance structure
     *              LRM instance structure
     *              PCM instance structure
     *              RPM instance structure
     *              SFV instance structure
     *              ZCM instance structure
     *              ZCS instance structure
     *              SRC instance structure
     *              Scaled samples buffer
     *              SRC input buffer
     *              SRC output buffer
     *              DLY instance structure
     */
    
    sz = sizeof(DI);
    sz += ast_query_mem_size();
    sz += bst_query_mem_size();
    sz += lrm_query_mem_size();
    sz += psc_query_mem_size();
    sz += rpm_query_mem_size();
    sz += sfv_query_mem_size();
    sz += zcm_query_mem_size();
    sz += zcs_query_mem_size();
    sz += src_query_mem_size(sample_rate, FEATURE_SAMPLE_RATE, MAX_FEATURE_BLOCK_SIZE, &src_max_input_size);
    sz += sizeof(float) * max_input_samples;            /* Scaled samples buffer */
    sz += sizeof(float) * src_max_input_size;           /* SRC input buffer */
    sz += sizeof(float) * MAX_FEATURE_BLOCK_SIZE;       /* SRC output buffer */

    /* Find the length of the delay line */
    delay_line_length = ((unsigned int) 2) * MAX(max_input_samples, src_max_input_size);
    sz += dly_query_mem_size(delay_line_length);        /* Delay line */
    
    return sz;
}

/*******************************************************************************************
int di_init(void *p_inst, size_t num_bytes, unsigned int sample_rate, unsigned int max_input_samples)
-Purpose:
    Initialize an instance of Dialogue Intelligence.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst                    - Pointer to instance memory.
    size_t num_bytes                - Size of memory pointed to by p_inst, in bytes.
    unsigned int sample_rate        - Sample rate, in Hz (either 32000, 44100 or 48000).
    unsigned int max_input_samples  - Maximum input block size, in samples.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int di_init(void *p_inst, size_t num_bytes, unsigned int sample_rate, unsigned int max_input_samples)
{
    unsigned int    src_max_input_size = 0;
    unsigned int    delay_line_length;
    DI              *p_di;
    size_t          sz_src;
    size_t          sz_dly;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG((sample_rate == 32000) || (sample_rate == 44100) || (sample_rate == 48000));
    ERR_CHECKARG(max_input_samples > 0);
    ERR_CHECKARG(num_bytes == di_query_mem_size(sample_rate, max_input_samples));
    
    /* Retrieve the instance pointer */
    p_di = (DI *) p_inst;
    
    /* Configure the instance pointers */
    sz_src = src_query_mem_size(sample_rate, FEATURE_SAMPLE_RATE, MAX_FEATURE_BLOCK_SIZE, &src_max_input_size);
    p_di->p_ast             = (((unsigned char *) p_di) + sizeof(DI));
    p_di->p_bst             = (((unsigned char *) p_di->p_ast) + ast_query_mem_size());
    p_di->p_lrm             = (((unsigned char *) p_di->p_bst) + bst_query_mem_size());
    p_di->p_psc             = (((unsigned char *) p_di->p_lrm) + lrm_query_mem_size());
    p_di->p_rpm             = (((unsigned char *) p_di->p_psc) + psc_query_mem_size());
    p_di->p_sfv             = (((unsigned char *) p_di->p_rpm) + rpm_query_mem_size());
    p_di->p_zcm             = (((unsigned char *) p_di->p_sfv) + sfv_query_mem_size());
    p_di->p_zcs             = (((unsigned char *) p_di->p_zcm) + zcm_query_mem_size());
    p_di->p_src             = (((unsigned char *) p_di->p_zcs) + zcs_query_mem_size());
    p_di->p_scaled_samples  = (float *) (((unsigned char *) p_di->p_src) + sz_src);
    p_di->p_src_in_data     = (float *) (((unsigned char *) p_di->p_scaled_samples) + (sizeof(float) * max_input_samples));
    p_di->p_src_out_data    = (float *) (((unsigned char *) p_di->p_src_in_data) + (sizeof(float) * src_max_input_size));    
    p_di->p_dly             = (((unsigned char *) p_di->p_src_out_data) + (sizeof(float) * MAX_FEATURE_BLOCK_SIZE));
    
    /* Setup each of the features */
    (void) ast_init(p_di->p_ast, ast_query_mem_size());
    (void) lrm_init(p_di->p_lrm, lrm_query_mem_size());
    (void) psc_init(p_di->p_psc, psc_query_mem_size());
    (void) rpm_init(p_di->p_rpm, rpm_query_mem_size());
    (void) sfv_init(p_di->p_sfv, sfv_query_mem_size());
    (void) zcm_init(p_di->p_zcm, zcm_query_mem_size());
    (void) zcs_init(p_di->p_zcs, zcs_query_mem_size());
    
    /* Initialize the boosting algorithm */
    (void) bst_init(p_di->p_bst, bst_query_mem_size());
    
    /* Initialize the SRC */
    (void) src_init(p_di->p_src, sz_src, MAX_FEATURE_BLOCK_SIZE, sample_rate, FEATURE_SAMPLE_RATE);
    src_max_input_size = src_get_input_length(p_di->p_src);
   
    /* Find the delay line length */
    delay_line_length = ((unsigned int) 2) * MAX(max_input_samples, src_max_input_size);
    
    /* Initialize the delay line */
    sz_dly = dly_query_mem_size(delay_line_length);
    (void) dly_init(p_di->p_dly, sz_dly, delay_line_length);
    
    /* Reset variables for block processing */
    p_di->max_block_count = INIT_DELAY;
    
    /* Reset the speech flag */
    p_di->is_speech = INVALID;

    return 0;

}

/*******************************************************************************************
int di_process(void *p_inst, float *p_samples, unsigned int num_samples)
-Purpose:
    Update Dialogue Intelligence for a block of audio.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    float *p_samples        - (in) Pointer to a block of audio.
    int num_samples         - (in) Number of samples contained in p_audio.  Required to be AST_BLOCK_SIZE.
-Returns:
    Returns the current classification: SPEECH, OTHER or INVALID.  INVALID is returned at
    the start of a program while samples are being collected to produce the first
    classification.
********************************************************************************************/
int di_process(void *p_inst, float *p_samples, unsigned int num_samples)
{
    unsigned int    i;
    float           tmp; 
    unsigned int    input_buffer_available;
    unsigned int    src_in_frame_size;
    float           max_sample;
    DI              *p_di;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_samples != NULL);
    ERR_CHECKARG(num_samples > 0);
    
    /* Retrieve the instance pointer */
    p_di = (DI *) p_inst;

    /* Scale and copy samples to the input buffer
     *
     * Dialogue Intelligence was developed and trained with PCM samples on the
     * range [-32768.0:+32768.0].  While it is normal for floating-point PCM samples
     * to be normalized to the range [-1.0:+1.0], the classifier configuration
     * depends on the samples falling on the larger range.  Hence Dialogue Intelligence
     * maintains the original range of [-32768.0:+32768.0].
     */
    for (i = 0; i < num_samples; i++)
    {
        p_di->p_scaled_samples[i] = 32768.0f * p_samples[i];
    }
    
    /* Put samples in circular buffer */
    dly_write_samples(p_di->p_dly, p_di->p_scaled_samples, num_samples);

    /* See how many samples are available to DI */
    input_buffer_available = dly_samples_available(p_di->p_dly);
                
    /* See how many samples we need to update DI */
    src_in_frame_size = src_get_input_length(p_di->p_src);

    /* Process until we don't have enough samples */
    while (src_in_frame_size <= input_buffer_available)
    {
        /* Get samples from the input delay buffer */
        dly_read_samples(p_di->p_dly, p_di->p_src_in_data, src_in_frame_size);

        /* Find the maximum sample */
        max_sample = 0.0f;
        for (i = 0; i < src_in_frame_size; i++)
        {
            tmp = (float) fabs(p_di->p_src_in_data[i]);
            if (tmp > max_sample)
            {
                max_sample = tmp;
            }
        }

        /* Force block to zero.  Gets read of sensitivity to low-energy frames. */
        if (max_sample < 1)
        {
            for (i = 0; i < src_in_frame_size; i++)
            {
                p_di->p_src_in_data[i] = 0.0f;
            }
        }

        /* Perform sample rate conversion */
        (void) src_process(p_di->p_src, src_in_frame_size, p_di->p_src_in_data, p_di->p_src_out_data);
        
        /* Update each of the features for this block */
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += AST_BLOCK_SIZE)
        {
            (void) ast_process_block(p_di->p_ast, &p_di->p_src_out_data[i], AST_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += LRM_BLOCK_SIZE)
        {
            (void) lrm_process_block(p_di->p_lrm, &p_di->p_src_out_data[i], LRM_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += PSC_BLOCK_SIZE)
        {
            (void) psc_process_block(p_di->p_psc, &p_di->p_src_out_data[i], PSC_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += RPM_BLOCK_SIZE)
        {
            (void) rpm_process_block(p_di->p_rpm, &p_di->p_src_out_data[i], RPM_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += SFV_BLOCK_SIZE)
        {
            (void) sfv_process_block(p_di->p_sfv, &p_di->p_src_out_data[i], SFV_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += ZCM_BLOCK_SIZE)
        {
            (void) zcm_process_block(p_di->p_zcm, &p_di->p_src_out_data[i], ZCM_BLOCK_SIZE);
        }
        for (i = 0; i < MAX_FEATURE_BLOCK_SIZE; i += ZCS_BLOCK_SIZE)
        {
            (void) zcs_process_block(p_di->p_zcs, &p_di->p_src_out_data[i], ZCS_BLOCK_SIZE);
        }
        
        /* Update the block counter */
        p_di->max_block_count++;

        /* Check whether it is time to update the frame */
        if (p_di->max_block_count == MAX_BLOCKS_PER_FRAME)
        {       
            /* Update each feature */
            p_di->p_feature_vector[AST_IDX] = ast_process_feature_frame(p_di->p_ast);
            p_di->p_feature_vector[LRM_IDX] = lrm_process_feature_frame(p_di->p_lrm);
            p_di->p_feature_vector[PSC_IDX] = psc_process_feature_frame(p_di->p_psc);
            p_di->p_feature_vector[RPM_IDX] = rpm_process_feature_frame(p_di->p_rpm);
            p_di->p_feature_vector[SFV_IDX] = sfv_process_feature_frame(p_di->p_sfv);
            p_di->p_feature_vector[ZCM_IDX] = zcm_process_feature_frame(p_di->p_zcm);
            p_di->p_feature_vector[ZCS_IDX] = zcs_process_feature_frame(p_di->p_zcs);

            /* Classify */
            p_di->is_speech = bst_classify(p_di->p_bst, p_di->p_feature_vector);

            /* Reset the block counter */
            p_di->max_block_count = 0;

        }

        /* See how many samples are available to DI now */
        input_buffer_available = dly_samples_available(p_di->p_dly);

        /* See how many samples we need to update DI again */
        src_in_frame_size = src_get_input_length(p_di->p_src);

    }
    
    return p_di->is_speech;
}
