/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Circular delay line

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include "dly.h"
#include "error.h"

/* State structure */
typedef struct
{
    float           *p_samples;     /* Pointer to audio samples */
    unsigned int    length;         /* Length of delay line */
    unsigned int    write_index;    /* Write index into p_samples */
    unsigned int    read_index;     /* Read index into p_samples */
} DLY;

/*******************************************************************************************
size_t dly_query_mem_size(unsigned int length)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    unsigned int length     - (in) Buffer length, in samples
-Returns:
    Returns the size of an instance, in bytes.
********************************************************************************************/
size_t dly_query_mem_size(unsigned int length)
{
    size_t sz;
    
    /* Memory is arranged as:
     * Address 0    DLY instance structure
     *              Internal floating point sample buffer 
     */
    sz = sizeof(DLY) + (sizeof(float) * length);
    return sz;
}

/*******************************************************************************************
int dly_init(void *p_inst, size_t num_bytes, unsigned int length)
-Purpose:
    Initialize an instance of DLY.    This function initializes an instance that has
    already been allocated by the calling code.  No memory allocation is performed
    by this function.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance
    size_t num_bytes    - (in) Size of memory pointed to by p_inst, in bytes.
    unsigned int length - (in) Buffer length, in samples
-Returns:
    0 on success, or -1 on error.
********************************************************************************************/
int dly_init(void *p_inst, size_t num_bytes, unsigned int length)
{
    DLY             *p_dly;
    unsigned int    i;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == dly_query_mem_size(length));
    ERR_CHECKARG(length > 0);

    p_dly = (DLY *) p_inst;

    /* Save the buffer length */
    p_dly->length = length;
    
    /* Initialize the samples buffer pointer */
    p_dly->p_samples = (float *) (((unsigned char *) p_inst) + sizeof(DLY));
    
    /* Zero the internal buffer */
    for (i = 0; i < p_dly->length; i++)
    {
        p_dly->p_samples[i] = 0.0f;
    }
    
    /* Reset the read and write indicies */
    p_dly->write_index = 0;
    p_dly->read_index = 0;
    
    return 0;
}

/*******************************************************************************************
void dly_write_samples(void *p_inst, float *p_samples, unsigned int sample_count)
-Purpose:
    Write samples into an instance of the DLY.
-Parameters:
    void *p_inst                - (in/out) Pointer to an instance
    float *p_samples            - (in) Pointer to array of pointers to samples
    unsigned int sample_count   - (in) Number of samples to write
-Returns:
    Nothing.
********************************************************************************************/
void dly_write_samples(void *p_inst, float *p_samples, unsigned int sample_count)
{
    unsigned int    i;
    unsigned int    write_index;
    DLY             *p_dly;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_samples != NULL);
    ERR_CHECKARG(sample_count > 0);
    ERR_CHECKARG(sample_count <= dly_space_available(p_inst));

    p_dly = (DLY *) p_inst;

    /* Copy into the buffer */
    write_index = p_dly->write_index;
    for (i = 0; i < sample_count; i++)
    {
        p_dly->p_samples[write_index] = p_samples[i];
        write_index++;
        write_index = (write_index == p_dly->length) ? 0 : write_index; 
    }

    /* Update the write index */
    p_dly->write_index = write_index;
    
    return;

}

/*******************************************************************************************
unsigned int dly_samples_available(void *p_inst)
-Purpose:
    Get the number of samples available for an instance of the DLY.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance
-Returns:
    Number of samples available.
********************************************************************************************/
unsigned int dly_samples_available(void *p_inst)
{
    int     number_available;
    DLY     *p_dly;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    p_dly = (DLY *) p_inst;
    
    /* Calculate the number of samples */
    number_available = ((int) p_dly->write_index) - ((int) p_dly->read_index);
    
    /* Unwrap the calculation, if necessary */
    if (number_available < 0)
    {
        number_available += p_dly->length;
    }
    
    return (unsigned int) number_available;
}

/*******************************************************************************************
unsigned int dly_space_available(void *p_inst)
-Purpose:
    Get the available space, in samples, for an instance of the DLY.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance
-Returns:
    Number of samples available.
********************************************************************************************/
unsigned int dly_space_available(void *p_inst)
{   
    DLY     *p_dly;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    p_dly = (DLY *) p_inst;
    
    return p_dly->length - dly_samples_available(p_dly);
}

/*******************************************************************************************
void dly_read_samples(void *p_inst, float *p_samples, unsigned int sample_count)
-Purpose:
    Read samples from an instance of the DLY
-Parameters:
    void *p_inst                - (in/out) Pointer to an instance
    float *p_samples            - (in) Pointer to array of pointers to samples
    unsigned int sample_count   - (in) Number of samples to read
-Returns:
    Nothing.
********************************************************************************************/
void dly_read_samples(void *p_inst, float *p_samples, unsigned int sample_count)
{
    unsigned int    i;
    unsigned int    read_index; 
    DLY             *p_dly;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_samples != NULL);
    ERR_CHECKARG(sample_count > 0);
    
    p_dly = (DLY *) p_inst;
    
    read_index = p_dly->read_index;
    for (i = 0; i < sample_count; i++)
    {
        p_samples[i] = p_dly->p_samples[read_index];
        read_index++;
        read_index = (read_index == p_dly->length) ? 0 : read_index; 
    }

    p_dly->read_index = read_index;

}

