/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Long rhythmic measure feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>

#include "lrm.h"
#include "fft.h"
#include "error.h"

/* Defines */
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / LRM_BLOCK_SIZE)
#define ACF_LENGTH              (LRM_BLOCK_SIZE/2)
#define FFT_SIZE                (LRM_BLOCK_SIZE*2)

/* Instance structure */
typedef struct
{
    float   p_power_estimates[BLOCKS_PER_FRAME];    /* Array of power estimates per block */
    int     power_estimate_idx;                     /* Index into p_power_estimates */
    float   p_old_power_estimates[ACF_LENGTH];      /* Array of prior power estimates */
    float   p_real[FFT_SIZE];                       /* "Real" buffer for FFT */
    float   p_imag[FFT_SIZE];                       /* "Imag" buffer for FFT */
} LRM;

/* Public functions */
/*******************************************************************************************
size_t lrm_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t lrm_query_mem_size(void)
{
    return sizeof(LRM);
}

/*******************************************************************************************
int lrm_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the long rythmic measure feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int lrm_init(void *p_inst, size_t num_bytes)
{
    int i;
    LRM *p_lrm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == lrm_query_mem_size());
    
    /* Retrieve the instance pointer */
    p_lrm = (LRM *) p_inst;

    /* Initialize the power estimate index */
    p_lrm->power_estimate_idx = 0;

    /* Initialize the power estimates */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_lrm->p_power_estimates[i] = 0.0f;
    }

    /* Initialize the old power estimates */
    for (i = 0; i < ACF_LENGTH; i++)
    {
        p_lrm->p_old_power_estimates[i] = 0.0f;
    }

    return 0;
}

/*******************************************************************************************
int lrm_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the long rythmic measure feature for a block of audio.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be LRM_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int lrm_process_block(void *p_inst, float *p_audio, int num_samples)
{
    int     i;
    float   power;
    LRM     *p_lrm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == LRM_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_lrm = (LRM *) p_inst;
        
    /* Calculate the short term power estimate */
    power = 0.0f;
    for (i = 0; i < LRM_BLOCK_SIZE; i++)
    {
        power += (p_audio[i] * p_audio[i]);
    }
    power /= (float) LRM_BLOCK_SIZE;
    
    /* Save the short-term power estimate */
    p_lrm->p_power_estimates[p_lrm->power_estimate_idx] = power;
    p_lrm->power_estimate_idx++;
    p_lrm->power_estimate_idx &= (BLOCKS_PER_FRAME - 1);
    
    return 0;

}

/*******************************************************************************************
float lrm_process_feature_frame(void *p_inst)
-Purpose:
    Calculate the long rythmic measure feature for a frame.
-Parameters:
    void *p_inst     - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float lrm_process_feature_frame(void *p_inst)
{
    float   return_value;
    int     power_estimate_idx;
    int     i;
    int     overlap;
    float   mean;
    float   max;
    float   *p_real;
    float   *p_imag;
    LRM     *p_lrm;
        
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_lrm = (LRM *) p_inst;
    
    /* Retrieve the real and imag pointers */
    p_real = p_lrm->p_real;
    p_imag = p_lrm->p_imag;
    
    /* Calculate mean of the current and old power estimates. */
    mean = 0.0f;
    for (i = 0; i < ACF_LENGTH; i++)
    {
        mean += p_lrm->p_power_estimates[i];
        mean += p_lrm->p_old_power_estimates[i];
    }
    mean /= (float) (BLOCKS_PER_FRAME * 2);

    /* The power estimates are held in a circular buffer.  power_estimate_idx
     * holds the index to the oldest estimate.
     */
    power_estimate_idx = p_lrm->power_estimate_idx;

    /* Pack the old and current power estimates into the "real" data array.  Zero pad to end.
     * Set the "imag" array to zero.
     */
    for (i = 0; i < ACF_LENGTH; i++)
    {
        p_real[i] = p_lrm->p_old_power_estimates[i] - mean;
        p_real[ACF_LENGTH + i] = p_lrm->p_power_estimates[power_estimate_idx] - mean;

        power_estimate_idx++;
        power_estimate_idx &= (ACF_LENGTH - 1);

        p_real[ACF_LENGTH * 2 + i] = 0.0f;
        p_real[ACF_LENGTH * 3 + i] = 0.0f;

        p_imag[i] = 0.0f;
        p_imag[ACF_LENGTH + i] = 0.0f;
        p_imag[ACF_LENGTH * 2 + i] = 0.0f;
        p_imag[ACF_LENGTH * 3 + i] = 0.0f;
    }
    
    /* Update the old power estimates */
    overlap = ACF_LENGTH - ACF_LENGTH / OVERLAP;
    power_estimate_idx = p_lrm->power_estimate_idx;
    
    for (i = 0; i < overlap; i++)
    {
        p_lrm->p_old_power_estimates[i] = p_lrm->p_old_power_estimates[i + ACF_LENGTH / OVERLAP];
    }
    for (i = 0; i < ACF_LENGTH / OVERLAP; i++)
    {
        p_lrm->p_old_power_estimates[i + overlap] = p_lrm->p_power_estimates[power_estimate_idx];

        power_estimate_idx++;
        power_estimate_idx &= (ACF_LENGTH - 1);
    }

    /* Find the ACF using an FFT. */

    /* FFT */
    (void) fftl(p_real, p_imag, FFT_SIZE);
    
    for (i = 0; i < FFT_SIZE; i++)
    {
        p_real[i] = p_imag[i] * p_imag[i] + p_real[i] * p_real[i];
        p_imag[i] = 0.0f;
    }
    
    /* IFFT */
    /* Should be conj(fft(conj(X)))/N but its real only so just an FFT */
    (void) fftl(p_real, p_imag, FFT_SIZE);

    /* Find the peak ACF */
    max = 0.0f;
    for (i = 30; i < (FFT_SIZE/2); i++)
    {
        max = ( max > p_real[i] ) ? max : p_real[i];
    }
    
    if (p_real[0] > 0.0f)
    {
        return_value = max / p_real[0];
    }
    else
    {
        return_value = 0.0f;
    }

    return return_value;
}
