/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Pause count metric feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>

#include "psc.h"
#include "error.h"

/* Defines */
#define POWER_RATIO             4.0f
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / PSC_BLOCK_SIZE)

/* Instance structure */
typedef struct
{
    float   p_st_power[BLOCKS_PER_FRAME];   /* Array of short term power estimates per block */
    int     st_power_idx;                   /* Index into p_st_power */
} PSC;

/* Public functions */
/*******************************************************************************************
size_t psc_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t psc_query_mem_size(void)
{
    return sizeof(PSC);
}

/*******************************************************************************************
int psc_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the pause count metric feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int psc_init(void *p_inst, size_t num_bytes)
{
    int     i;
    PSC     *p_psc;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == psc_query_mem_size());

    /* Retrieve the instance pointer */
    p_psc = (PSC *) p_inst;

    /* Initialize the short term power index */
    p_psc->st_power_idx = 0;
    
    /* Initialize the short term powers */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_psc->p_st_power[i] = 0.0f;
    }

    return 0;
}

/*******************************************************************************************
int psc_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the pause count metric feature for a block of audio.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be PSC_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int psc_process_block(void *p_inst, float *p_audio, int num_samples)
{
    int     i;
    float   power;
    PSC     *p_psc;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == PSC_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_psc = (PSC *) p_inst;

    /* Calculate the short term power estimate */
    power = 0.0f;
    for (i = 0; i < PSC_BLOCK_SIZE; i++)
    {
        power += (p_audio[i] * p_audio[i]);
    }
    power /= (float) PSC_BLOCK_SIZE;

    /* Save the short term power estimate */
    p_psc->p_st_power[p_psc->st_power_idx] = power;
    p_psc->st_power_idx++;
    p_psc->st_power_idx &= (BLOCKS_PER_FRAME - 1);

    return 0;

}

/*******************************************************************************************
float psc_process_feature_frame(void *p_inst)
-Purpose:
    Calculate the pause count metric feature for a frame of audio.
-Parameters:
    void *p_inst     - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float psc_process_feature_frame(void *p_inst)
{
    float       return_value;
    int         i;
    int         pause_count;
    float       frame_power;
    PSC         *p_psc;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_psc = (PSC *) p_inst;
      
    /* Find the long term variance */
    frame_power = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        frame_power += p_psc->p_st_power[i];
    }
    frame_power /= (float) (BLOCKS_PER_FRAME);

    /* Find the pause count.  The pause count does not update on the first
     * block of audio.
     */
    pause_count = 0;
    for (i = 1; i < BLOCKS_PER_FRAME; i++)
    {
        if ((frame_power > p_psc->p_st_power[i] * POWER_RATIO) &&
            (frame_power < p_psc->p_st_power[i - 1] * POWER_RATIO)
            )
        {
            pause_count++;
        }
    }
    
    /* Convert the pause count to a float */
    return_value = (float) pause_count;

    return return_value;
}
