/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Rhythmic measure feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>

#include "rpm.h"
#include "fft.h"
#include "error.h"

/* Defines */
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / RPM_BLOCK_SIZE)
#define FFT_SIZE                RPM_BLOCK_SIZE
#define ACF_LENGTH              (FFT_SIZE/2)

/* Instance structure */
typedef struct
{
    float   p_st_variance[BLOCKS_PER_FRAME];    /* Short term variances per block */
    int     st_variance_idx;                    /* Array into p_st_variance */
    float   p_old_acf[ACF_LENGTH];              /* Array of prior autocorrelations */
    float   p_older_acf[ACF_LENGTH];            /* Array of prior autocorrelations */
    float   p_real[FFT_SIZE];                   /* "Real" buffer for FFT */
    float   p_imag[FFT_SIZE];                   /* "Imag" buffer for FFT */
} RPM;

/* Public prototypes */
/*******************************************************************************************
size_t rpm_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t rpm_query_mem_size(void)
{
    return sizeof(RPM);
}

/*******************************************************************************************
int rpm_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the rythmic measure feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int rpm_init(void *p_inst, size_t num_bytes)
{
    int i;
    RPM *p_rpm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == rpm_query_mem_size());
    
    /* Retrieve the instance pointer */
    p_rpm = (RPM *) p_inst;

    /* Initialize the short term variance index */
    p_rpm->st_variance_idx = 0;

    /* Initialize the short term variances */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_rpm->p_st_variance[i] = 0.0f;
    }

    /* Initialize the autocorrelations */
    for (i = 0; i < ACF_LENGTH; i++)
    {
        p_rpm->p_old_acf[i] = 0.0f;
        p_rpm->p_older_acf[i] = 0.0f;
    }

    return 0;
}

/*******************************************************************************************
int rpm_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the rythmic measure feature for a block of audio.
-Parameters:
    void *p_inst          - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be RPM_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int rpm_process_block(void *p_inst, float *p_audio, int num_samples)
{
    int     i;
    float   variance;
    float   mean;
    float   temp;
    RPM     *p_rpm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == RPM_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_rpm = (RPM *) p_inst;

    /* Calculate short term mean */
    mean = 0.0f;
    for (i = 0; i < RPM_BLOCK_SIZE; i++)
    {
        mean += p_audio[i];
    }
    mean /= (float) RPM_BLOCK_SIZE;

    /* Calculate short term variance estimate */
    variance = 0.0f;
    for (i = 0; i < RPM_BLOCK_SIZE; i++)
    {
        temp = p_audio[i] - mean;
        temp *= temp;
        variance += temp;
    }
    variance /= (float) RPM_BLOCK_SIZE;

    /* Save the short-term variance estimate */
    p_rpm->p_st_variance[p_rpm->st_variance_idx] = variance;
    p_rpm->st_variance_idx++;
    p_rpm->st_variance_idx &= (BLOCKS_PER_FRAME - 1);

    return 0;

}

/*******************************************************************************************
float rpm_process_feature_frame(void *p_inst)
-Purpose:
    Calculate the rythmic measure feature for a frame.
-Parameters:
    void *p_inst  - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float rpm_process_feature_frame(void *p_inst)
{
    float   return_value;
    int     i;
    int     st_variance_idx;
    float   mean;
    float   *p_real;
    float   *p_imag;
    float   long_variance;
    float   max;
    float   temp;
    RPM     *p_rpm;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_rpm = (RPM *) p_inst;

    /* Retrieve the real and imag pointers */
    p_real = p_rpm->p_real;
    p_imag = p_rpm->p_imag;

    /* Find the long variance */
    long_variance = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        long_variance += p_rpm->p_st_variance[i];
    }
    long_variance /= (float) (BLOCKS_PER_FRAME);

    if (long_variance > 0.0f)
    {
        /* Calculate mean */
        mean = 0.0f;
        for (i = 0; i < ACF_LENGTH; i++)
        {
            mean += p_rpm->p_st_variance[i] / long_variance;
        }
        mean /= (float) BLOCKS_PER_FRAME;

        /* The short-term varaince estimates are held in a circular buffer.  st_variance_idx
         * holds the index to the oldest estimate.
         */
        st_variance_idx = p_rpm->st_variance_idx;
        
        /* Find the ACF using an FFT. */
        for (i = 0; i < ACF_LENGTH; i++)
        {
            p_real[i] = (p_rpm->p_st_variance[st_variance_idx] / long_variance) - mean;
            st_variance_idx++;
            st_variance_idx &= (ACF_LENGTH - 1);

            p_imag[i] = 0.0f;
            p_real[ACF_LENGTH + i] = 0.0f;
            p_imag[ACF_LENGTH + i] = 0.0f;
        }

        /* FFT */
        (void) fftl(p_real, p_imag, FFT_SIZE);
        
        for (i = 0; i < FFT_SIZE; i++)
        {
            p_real[i] = p_imag[i] * p_imag[i] + p_real[i] * p_real[i];
            p_imag[i] = 0.0f;
        }
        
        /* IFFT */
        /* Should be conj(fft(conj(X)))/N but its real only so just an FFT */
        (void) fftl(p_real, p_imag, FFT_SIZE);
        
        /* Add scaled verions of the prior (old and older) ACF's.  Update the prior ACF's. */
        for (i = 0; i < (FFT_SIZE/2); i++)
        {
            temp = p_real[i];
            p_real[i] += (p_rpm->p_old_acf[i] / 2.0f);
            p_real[i] += (p_rpm->p_older_acf[i] / 4.0f);
            p_rpm->p_older_acf[i] = p_rpm->p_old_acf[i];
            p_rpm->p_old_acf[i] = temp;
        }
        
        /* Find the peak ACF. */
        max = 0.0f;
        for(i = 10; i < (FFT_SIZE/2); i++)
        {
            max = ( max > p_real[i] ) ? max : p_real[i];
        }
        if (p_real[0] != 0.0f)
        {
            return_value = max / p_real[0];
        }
        else
        {
            return_value = 0.0f;
        }
    }
    else
    {
        return_value = 0.0f;
    }

    return return_value;
}
