/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Spectral flux feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>

#include "sfv.h"
#include "fft.h"
#include "error.h"

/* Defines */
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / SFV_BLOCK_SIZE)
#define FFT_SIZE                SFV_BLOCK_SIZE
#define LOWER_LIMT              6
#define UPPER_LIMIT             500

/* Instance structure */
typedef struct
{
    float   p_l2_norm[BLOCKS_PER_FRAME];    /* Array of l2-norms, per block */
    int     l2_norm_idx;                    /* Index into p_l2_norm */
    float   p_old_spectral_mag[FFT_SIZE/2]; /* Array of old spectral magnitudes */
    float   p_real[FFT_SIZE];               /* "Real" buffer for FFT */
    float   p_imag[FFT_SIZE];               /* "Imag" buffer for FFT */
} SFV;

/* Externs */
extern const float p_sfv_hamming[];

/* Public functions */
/*******************************************************************************************
size_t sfv_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t sfv_query_mem_size(void)
{
    return sizeof(SFV);
}

/*******************************************************************************************
int sfv_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the weighted spectral flux feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int sfv_init(void *p_inst, size_t num_bytes)
{
    int i;
    SFV *p_sfv;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == sfv_query_mem_size());
    
    /* Retrieve the instance pointer */
    p_sfv = (SFV *) p_inst;

    /* Initialize the l2-norm index */
    p_sfv->l2_norm_idx = 0;

    /* Initialize the l2-norms */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_sfv->p_l2_norm[i] = 0.0f;
    }

    /* Initialize the old spectral magnitudes */
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        p_sfv->p_old_spectral_mag[i] = 0.0f;
    }

    return 0;
}

/*******************************************************************************************
int sfv_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the weighted spectral flux feature for a block of audio.
-Parameters:
    void *p_inst          - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be SFV_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int sfv_process_block(void *p_inst, float *p_audio, int num_samples)
{
    float       *p_real;
    float       *p_imag;
    int         i;
    float       weight;
    float       max;
    float       norm;
    float       temp;
    SFV         *p_sfv;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == SFV_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_sfv = (SFV *) p_inst;
    
    /* Retrieve the real and imag pointers */
    p_real = p_sfv->p_real;
    p_imag = p_sfv->p_imag;

    /* Normalize the audio data, apply the hamming window and populate the real/imaginary buffers */
    max = 1.0f;
    for (i = 0; i < SFV_BLOCK_SIZE; i++)
    {
        max = (max > (float) (fabs(p_audio[i]))) ? max : (float) (fabs(p_audio[i]));
    }
    for (i = 0; i < FFT_SIZE; i++)
    {
        temp = (p_audio[i]) / max;
        temp *= p_sfv_hamming[i];
        p_real[i] = temp;
        p_imag[i] = 0.0f;
    }

    /* Calculate FFT */
    (void) fftl(p_real, p_imag, FFT_SIZE);

    /* Calculate sqrt( x .* conj(x) ) */
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        p_real[i] = p_real[i] * p_real[i] + p_imag[i] * p_imag[i];
        if (p_real[i] > 0.0f)
        {
            p_real[i] = (float) (sqrt(p_real[i]));
        }
        else
        {
            p_real[i] = 0.0f;
        }
    }

    /* Calculate weighting */
    weight = 0.0f;
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        weight += (p_real[i] * p_real[i] + p_sfv->p_old_spectral_mag[i] * p_sfv->p_old_spectral_mag[i]);
    }
    weight /= (float) ((FFT_SIZE/2) * 2);
    if (weight > 0.0f)
    {
        weight = (float) (sqrt(weight));
    }
    else
    {
        weight = 1.0f;
    }

    /* Calculate spectral flux */
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        temp = p_real[i];
        if (i > LOWER_LIMT && i < UPPER_LIMIT)
        {
            p_real[i] -= p_sfv->p_old_spectral_mag[i];
        }
        else
        {
            p_real[i] = 0.0f;
        }

        p_sfv->p_old_spectral_mag[i] = temp;
        p_real[i] /= weight;

    }

    /* Calculate the l2 norm of the weighted spectral flux */
    norm = 0.0f;
    for (i = 0; i < (FFT_SIZE/2); i++)
    {
        p_real[i] = (float) (fabs(p_real[i]));
        norm += (p_real[i] * p_real[i]);
    }

    /* Save the l2-norm */
    p_sfv->p_l2_norm[p_sfv->l2_norm_idx] = norm;
    p_sfv->l2_norm_idx++;
    p_sfv->l2_norm_idx &= (BLOCKS_PER_FRAME - 1);

    return 0;
}

/*******************************************************************************************
float sfv_process_feature_frame(void *p_inst)
-Purpose:
    Calculate the weighted spectral flux feature for a frame.
-Parameters:
    void *p_inst  - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float sfv_process_feature_frame(void *p_inst)
{
    float   return_value;    
    int     i;
    SFV     *p_sfv;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_sfv = (SFV *) p_inst;

    /* Calculate the feature */
    return_value = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        return_value += p_sfv->p_l2_norm[i];
    }
    return_value /= (float) BLOCKS_PER_FRAME;
    if (return_value > 0.0f)
    {
        return_value = (float) (log(return_value));
    }
    else
    {
        return_value = 0.0f;
    }


    return return_value;

}

