/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Sample rate converter

    Date: $Date$

    Revision: $Revision$

 ***************************************************************************/

/* Includes */
#include <stddef.h>
#include <math.h>

#include "src.h"
#include "error.h"

/* Defines */
#define RATIO_TABLE_LENGTH      3

/* State structure */
typedef struct
{
    unsigned int up_factor;              /* Up sample ratio */
    unsigned int down_factor;            /* Down sample ratio */
    unsigned int output_buffer_size;     /* Size of output bufer */
    unsigned int input_size;             /* Size of input buffer */
    unsigned int filter_length;          /* Length of decimation filter */
    unsigned int circular_buffer_size;   /* Size of decimated sample buffer */
    unsigned int circular_mask;          /* Mask for circular buffer index */
    unsigned int filter_commutator;      /* Commutator state */
    unsigned int circular_buffer_ptr;    /* Circular buffer pointer */
    unsigned int input_data_remaining;   /* Data remaining in the input buffer */
    const float  *p_filter_coefficients; /* Decimation filter coefficients */
    float        *p_circular_buffer;     /* Decimated sample buffer */
    float        saved_input;            /* Saved input */
    int          init;                   /* Take into account for initialization */
} SRC;

/* Configuration structure */
typedef struct
{
    unsigned int fs_in;          /* Input rate */
    unsigned int fs_out;         /* Output rate */
    unsigned int up_factor;      /* Up conversion factor */
    unsigned int down_factor;    /* Down conversion factor */
    unsigned int filt_order;     /* Filter order */
} SRC_RATIO_TABLE;

/* Ratio tables, for down-sampling only */
static const SRC_RATIO_TABLE p_src_ratio_table[RATIO_TABLE_LENGTH] =
{
/*    +- fs_in
 *    |      +- fs_out
 *    |      |      +- up_factor
 *    |      |      |    +- down_factor
 *    |      |      |    |    +- filt_order       
 *    |      |      |    |    |         */
    { 32000, 16000, 1,   2,   SRC_FILT_LEN_32_16  },
    { 44100, 16000, 160, 441, SRC_FILT_LEN_441_16 },
    { 48000, 16000, 1,   3,   SRC_FILT_LEN_48_16  }
};

/* Externs */
extern const float p_src_filter_coefs_48_16[];
extern const float p_src_filter_coefs_441_16[];
extern const float p_src_filter_coefs_32_16[];

/* Private functions */
static int calculate_src_params(unsigned int fs_in, unsigned int fs_out, unsigned int output_size, unsigned int *up_factor,
                                unsigned int *down_factor, unsigned int *filt_order, unsigned int *circ_buff_size,
                                unsigned int *max_input_size);

/* Public functions */
/*******************************************************************************************
size_t src_query_mem_size(unsigned int fs_in, unsigned int fs_out, unsigned int output_size, unsigned int *max_input_size)
-Purpose:
    Get the size of an instance, in bytes, for a specified input and output sampling rate.
-Parameters:
    unsigned int fs_in           - (in) Input sampling rate, in Hz.  Must be set to 32000, 44100 or 48000.
    unsigned int fs_out          - (in) Output sampling rate, in Hz.  Must be set to 16000.
    unsigned int output_size     - (in) Length of output buffer, in samples
    unsigned int *max_input_size - (out) Length of input buffer, in samples, based on specified configuration
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t src_query_mem_size(unsigned int fs_in, unsigned int fs_out, unsigned int output_size, unsigned int *max_input_size)
{
    size_t          sz;    
    unsigned int    up_factor;
    unsigned int    down_factor;
    unsigned int    filt_order;
    unsigned int    circular_buffer_size;
     
    /* Calculate the circular buffer size and and maximum input size */
    if (calculate_src_params(fs_in, fs_out, output_size, &up_factor, &down_factor, &filt_order, &circular_buffer_size, max_input_size) < 0)
    {
        /* Error */
        return 0;
    }
        
    /* Memory is arranged as:
     * Address 0    SRC instance structure
     *              Array of "circular_buffer_size" floating point samples
     */    
    sz = sizeof(SRC) + (sizeof(float) * circular_buffer_size);
    return sz;
}

/*******************************************************************************************
int src_init(void *p_inst, size_t num_bytes, unsigned int output_buffer_size, unsigned int fs_in, unsigned int fs_out)
-Purpose:
    Initialize an instance of SRC.  This function initializes an instance that has
    already been allocated by the calling code.  No memory allocation is performed
    by this function.
-Parameters:
    void *p_inst                    - (in/out) Pointer to an instance
    size_t num_bytes                - (in) Size of memory pointed to by p_inst, in bytes.
    unsigned int output_buffer_size - (in) Buffer length, in samples
    unsigned int fs_in              - (in) Input sampling rate, in Hz.  Must be set to 32000, 44100 or 48000.
    unsigned int fs_out             - (in) Output sampling rate, in Hz.  Must be set to 16000.
-Returns:
    0 on success, or -1 on error.
********************************************************************************************/
int src_init(void *p_inst, size_t num_bytes, unsigned int output_buffer_size, unsigned int fs_in, unsigned int fs_out)
{
    unsigned int    up_factor;
    unsigned int    down_factor;
    unsigned int    filt_order;
    unsigned int    i;
    unsigned int    dummy = 0;
    SRC             *p_src;
    
    /* Prevent "unused variable" warnings when the ERR_CHECKARG macro is disabled */
    (void) dummy;
    
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(output_buffer_size > 0);
    ERR_CHECKARG((fs_in == 32000) || (fs_in == 44100) || (fs_in == 48000));
    ERR_CHECKARG(fs_out == 16000);
    ERR_CHECKARG(num_bytes == src_query_mem_size(fs_in, fs_out, output_buffer_size, &dummy));
    
    /* Retrieve the instance pointer */
    p_src = (SRC *) p_inst;

    /* Calculate the SRC parameters */
    if (calculate_src_params(fs_in, fs_out, output_buffer_size, &up_factor, &down_factor, &filt_order, &p_src->circular_buffer_size, &p_src->input_size) < 0)
    {
        /* Error */
        return -1;
    }

    /* Find the filter coeffients array */
    switch (fs_in)
    {
    case 48000:
        p_src->p_filter_coefficients = p_src_filter_coefs_48_16;
        break;
        
    case 44100:
        p_src->p_filter_coefficients = p_src_filter_coefs_441_16;
        break;
        
    case 32000:
        p_src->p_filter_coefficients = p_src_filter_coefs_32_16;
        break;
        
    default:
        /* Should never get here due to checks above */
        return -1;
    }
    
    /* Find the circular buffer mask */
    p_src->circular_mask = p_src->circular_buffer_size - 1;

    /* Find the circular buffer and clear it */
    p_src->p_circular_buffer = (float *) ( ((unsigned char *) p_inst) + sizeof(SRC) );
    for (i = 0; i < p_src->circular_buffer_size; i++)
    {
        p_src->p_circular_buffer[i] = 0.0f;
    }
     
    /* Populate the SRC struct */
    p_src->up_factor = up_factor;
    p_src->down_factor = down_factor;
    p_src->output_buffer_size = output_buffer_size;
    p_src->filter_length = filt_order;
    p_src->init = 0;
    p_src->filter_commutator = 0;
    p_src->circular_buffer_ptr = 0;
    p_src->input_data_remaining = 0;
    p_src->saved_input = 0.0;
       
    return 0;
}

/*******************************************************************************************
int src_process(void *p_inst, unsigned int input_length, float *p_input, float *p_output)
-Purpose:
    Perform sample rate conversion on a block of audio samples.
-Parameters:
    void *p_inst              - (in/out) Pointer to an instance
    unsigned int input_length - (in) Number of elements in array pointed to by p_input
    float *p_input            - (in) Input samples
    float *p_output           - (out) Output samples
-Returns:
    0 on success, or -1 on error.
********************************************************************************************/
int src_process(void *p_inst, unsigned int input_length, float *p_input, float *p_output)
{
    const float     *p_filter_coefficients;
    unsigned int    output_size;
    unsigned int    filter_length;
    unsigned int    up_factor;
    unsigned int    down_factor;
    unsigned int    init_offset;
    unsigned int    circular_mask;
    unsigned int    temp_ptr;
    float           conv_sum;
    float           *p_circular_buffer;
    unsigned int    circular_buffer_ptr;
    unsigned int    filter_commutator;
    unsigned int    input_data_remaining;
    unsigned int    input_buffer_ptr;
    unsigned int    i, j;
    SRC             *p_src;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_input != NULL);
    ERR_CHECKARG(p_output != NULL);
    ERR_CHECKARG(input_length > 0);
    ERR_CHECKARG(input_length == src_get_input_length(p_inst));
    
    /* Retrieve the instance pointer */
    p_src = (SRC *) p_inst;

    /* Initialize the offset */
    init_offset = 0;
    if (p_src->init == 0)
    {
        p_src->circular_buffer_ptr = 0;
        p_src->p_circular_buffer[0] = *p_input;
            
        init_offset = 1;
        p_src->init++;
    }

    /* Copy certain parameters from the instance to local storage, for
     * speed and readability.
     */
    p_filter_coefficients   = p_src->p_filter_coefficients;
    output_size             = p_src->output_buffer_size;
    filter_length           = p_src->filter_length;
    up_factor               = p_src->up_factor;
    down_factor             = p_src->down_factor;
    circular_mask           = p_src->circular_mask;
    p_circular_buffer       = p_src->p_circular_buffer;
    circular_buffer_ptr     = p_src->circular_buffer_ptr;
    filter_commutator       = p_src->filter_commutator;
    input_data_remaining    = p_src->input_data_remaining;
    input_buffer_ptr        = init_offset;

    /* Perform sample rate conversion */
    for (i = 0; i < output_size; i++)
    {
        /* Compute convolution sum */
        conv_sum = 0;
        temp_ptr = circular_buffer_ptr;
        for (j = filter_commutator; j < filter_length; j += up_factor)
        {
            conv_sum += (p_filter_coefficients[j]) * (p_circular_buffer[temp_ptr]);
            temp_ptr--;
            temp_ptr &= circular_mask;
        }
        
        /* Save the output */
        p_output[i] = conv_sum;

        /* Update the commutator */
        filter_commutator += down_factor;
        while (filter_commutator > up_factor)
        {
            filter_commutator -= up_factor;

            circular_buffer_ptr++;
            circular_buffer_ptr &= circular_mask;

            if (input_data_remaining > 0)
            {
                p_circular_buffer[circular_buffer_ptr] = p_src->saved_input;
                input_data_remaining = 0;
            }
            else
            {
                p_circular_buffer[circular_buffer_ptr] = p_input[input_buffer_ptr];
                input_buffer_ptr++;
            }
        }
    }

    if (input_buffer_ptr == input_length)
    {
        input_data_remaining = 0;
    }
    else
    {
        input_data_remaining = 1;
        p_src->saved_input = p_input[input_buffer_ptr];
    }
    
    /* Copy local variables back to the instance */
    p_src->circular_buffer_ptr  = circular_buffer_ptr;
    p_src->filter_commutator    = filter_commutator;
    p_src->input_data_remaining = input_data_remaining;

    return 0;
}

/*******************************************************************************************
unsigned int src_get_input_length(void *p_inst)
-Purpose:
    Computes the number of samples needed by the sample rate converter.
-Parameters:
    void *p_inst    - (in/out) Pointer to an instance
-Returns:
    Number of input samples required by the SRC.
********************************************************************************************/
unsigned int src_get_input_length(void *p_inst)
{
    SRC     *p_src;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Reteive the instance pointer */
    p_src = (SRC *) p_inst;
    
    return (p_src->input_size - p_src->input_data_remaining);
}


/* Private functions */
/*******************************************************************************************
static int calculate_src_params(unsigned int fs_in, unsigned int fs_out, unsigned int output_size, unsigned int *up_factor,
                                unsigned int *down_factor, unsigned int *filt_order, unsigned int *circ_buff_size,
                                unsigned int *max_input_size)
-Purpose:
    Calculate the parameters of the SRC that are required for calculating memory.
-Parameters:
    unsigned int fs_in           - (in) Input sampling rate, in Hz.  Must be set to 32000, 44100 or 48000.
    unsigned int fs_out          - (in) Output sampling rate, in Hz.  Must be set to 16000.
    unsigned int output_size     - (in) Length of output buffer, in samples
    unsigned int *up_factor      - (out) Upsampling factor
    unsigned int *down_factor    - (out) Downsampling factor
    unsigned int *filt_order     - (out) Filter order
    unsigned int *circ_buff_size - (out) Size of the circular buffer
    unsigned int *max_input_size - (out) Length of input buffer, in samples, based on specified configuration
-Returns:
    Returns 0 on success, or -1 on error.
********************************************************************************************/
static int calculate_src_params(unsigned int fs_in, unsigned int fs_out, unsigned int output_size, unsigned int *up_factor,
                                unsigned int *down_factor, unsigned int *filt_order, unsigned int *circ_buff_size,
                                unsigned int *max_input_size)
{
    unsigned int i;
    unsigned int temp;

    /* Find the sample rate conversion factors */
    i = 0;
    *up_factor = 0;
    *down_factor = 0;
    *filt_order = 0;
    while ((*up_factor == 0) && (i <= RATIO_TABLE_LENGTH))
    {
        if (fs_in == p_src_ratio_table[i].fs_in && fs_out == p_src_ratio_table[i].fs_out)
        {
            *up_factor   = p_src_ratio_table[i].up_factor;
            *down_factor = p_src_ratio_table[i].down_factor;
            *filt_order  = p_src_ratio_table[i].filt_order;
        }
        i++;
    }
    if (*up_factor == 0)
    {
        /* Sample rate conversion factors not found! */
        return -1;
    }
    
    /* Compute the size of input buffer */
    if ((*down_factor % *up_factor) > 0)
    {
        *max_input_size = ((output_size * *down_factor) / *up_factor) + 1;
    }
    else
    {
        *max_input_size = ((output_size * *down_factor) / *up_factor);
    }
    
    /* Compute the size of circular buffer */
    if ((*filt_order % *up_factor) > 0)
    {
        temp = (*filt_order / *up_factor) + 1;
    }
    else
    {
        temp = *filt_order / *up_factor;
    }
    
    /* Find the circular buffer size */
    for (i = 1; i < temp; i <<= 1);
    /* Note: for loop has no body */
    
    *circ_buff_size = i;
    
    return 0;

}
