/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Zero crossing mean to median ratio feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "zcm.h"
#include "error.h"

/* Defines */
#define BLOCKS_PER_FRAME            (FEATURE_FRAME_SIZE / ZCM_BLOCK_SIZE)
#define POSITIVE                    0
#define NEGATIVE                    1

/* Instance structure */
typedef struct
{
    float   p_crossing_count[BLOCKS_PER_FRAME]; /* Crossing counts per block */
    int     crossing_count_idx;                 /* Index into p_crossing_count */
    int     crossing_state;                     /* Current crossing state */
    float   p_sorting_buffer[BLOCKS_PER_FRAME]; /* Temp buffer for sorting p_crossing_count */
} ZCM;

/* Private prototypes */
static void sort(float *p_data, int size);

/* Public functions */
/*******************************************************************************************
size_t zcm_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t zcm_query_mem_size(void)
{
    return sizeof(ZCM);
}

/*******************************************************************************************
int zcm_init(void *p_inst)
-Purpose:
    Initialize the zero crossing metric feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int zcm_init(void *p_inst, size_t num_bytes)
{
    int i;
    ZCM *p_zcm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == zcm_query_mem_size());

    /* Retrieve the instance pointer */
    p_zcm = (ZCM *) p_inst;

    /* Initialize the crossing count index */
    p_zcm->crossing_count_idx = 0;

    /* Initialize the crossing count */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_zcm->p_crossing_count[i] = 0.0f;
    }

    /* Initialize the crossing state */
    p_zcm->crossing_state = POSITIVE;

    return 0;
}


/*******************************************************************************************
int zcm_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the zero crossing metric feature for a block of audio.
-Parameters:
    void *p_inst          - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be ZCM_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int zcm_process_block(void *p_inst, float *p_audio, int num_samples)
{
    int     i;
    int     cross_counter;
    int     *p_crossing_state;
    ZCM     *p_zcm;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == ZCM_BLOCK_SIZE);

    /* Retrieve the instance pointer */
    p_zcm = (ZCM *) p_inst;

    /* Get extra payload info */
    p_crossing_state = &(p_zcm->crossing_state);

    /* Calculate the zero crossing rate */
    cross_counter = 0;
    for (i = 0; i < ZCM_BLOCK_SIZE; i++)
    {
        if ((*p_crossing_state == NEGATIVE) && 
            (p_audio[i] > 0.0f)
            )
        {
            /* Positive going signal */
            *p_crossing_state = POSITIVE;
            cross_counter++;
        }
        else if ((*p_crossing_state == POSITIVE) &&
                 (p_audio[i] < 0.0f)
                 )
        {
            /* Negative going signal */
            *p_crossing_state = NEGATIVE;
            cross_counter++;
        }
    }

    /* Save the crossing count */
    p_zcm->p_crossing_count[p_zcm->crossing_count_idx] = (float) cross_counter;
    p_zcm->crossing_count_idx++;
    p_zcm->crossing_count_idx &= (BLOCKS_PER_FRAME - 1);
    
    return 0;
}

/*******************************************************************************************
float zcm_process_feature_frame(void *p_inst, size_t num_bytes)
-Purpose:
    Calculate the zero crossing metric feature for a frame.
-Parameters:
    void *p_inst  - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float zcm_process_feature_frame(void *p_inst)
{
    float   return_value;
    int     i;
    float   mean;
    float   median;
    float   *p_sorting_buffer;
    ZCM     *p_zcm;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_zcm = (ZCM *) p_inst;

    /* Retrieve the pointer to the sorting buffer */
    p_sorting_buffer = p_zcm->p_sorting_buffer;

    /* Calculate the mean crossing rate */
    mean = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        mean += (p_zcm->p_crossing_count[i]);
        p_sorting_buffer[i] = p_zcm->p_crossing_count[i];
    }
    mean /= (float) (BLOCKS_PER_FRAME);
    
    /* Sort the zero crossing rates */
    sort(p_sorting_buffer, BLOCKS_PER_FRAME);
    
    /* Find the median crossing rate */
    median = (p_sorting_buffer[BLOCKS_PER_FRAME / 2 - 1] + p_sorting_buffer[BLOCKS_PER_FRAME / 2]) / 2.0f;
    
    /* Find the mean to median ratio */
    if (mean != 0.0f)
    {
        return_value = median / mean;
    }
    else
    {
        return_value = 0.0f;
    }

    return return_value;
}

/* Private functions */
/*******************************************************************************************
static void sort(float *p_data, int size)
-Purpose:
    Sort an array in place.
-Parameters:
    float *p_data       - (in/out) Pointer to the array to sort
    int size            - (in) Length of the array
-Returns:
    Nothing.
********************************************************************************************/
static void sort(float *p_data, int size)
{
    int     n;
    int     k;
    int     index;
    float   max;
    
    for (n = 0; n < size - 1; n++)
    {
        max = -1e36f;
        index = 0;
        for (k = n; k < size; k++)
        {
            if (p_data[k] > max)
            {
                max = p_data[k];
                index = k;
            }
        }
        p_data[index] = p_data[n];
        p_data[n] = max;
    }
}
