/****************************************************************************
    This program is protected under international and U.S. copyright laws
    as an unpublished work. This program is confidential and proprietary to
    Dolby Laboratories. Reproduction or disclosure, in whole or in part, or
    the production of derivative works therefrom without the express
    permission of Dolby Laboratories is prohibited.
    Copyright 2003-2011 by Dolby Laboratories. All rights reserved.

    Description: Zero crossing skew feature

    Date: $Date$

    Revision: $Revision$

***************************************************************************/

/* Includes */
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "zcs.h"
#include "error.h"

/* Defines */
#define BLOCKS_PER_FRAME        (FEATURE_FRAME_SIZE / ZCS_BLOCK_SIZE)
#define POSITIVE                0
#define NEGATIVE                1

/* Instance structure */
typedef struct
{
    float   p_crossing_count[BLOCKS_PER_FRAME]; /* Crossing counts per block */
    int     crossing_count_idx;                 /* Index into p_crossing_count */
    int     crossing_state;                     /* Current crossing state */
} ZCS;

/* Public functions */
/*******************************************************************************************
size_t zcs_query_mem_size(void)
-Purpose:
    Get the size of an instance, in bytes.
-Parameters:
    None.
-Returns:
    Returns the size of an instance, in bytes, or 0 on error.
********************************************************************************************/
size_t zcs_query_mem_size(void)
{
    return sizeof(ZCS);
}

/*******************************************************************************************
int zcs_init(void *p_inst, size_t num_bytes)
-Purpose:
    Initialize the zero crossing skew feature structure.  This function initializes an
    instance that has already been allocated by the calling code.  No memory allocation
    is performed by this function.
-Parameters:
    void *p_inst            - (in/out) Pointer to an instance.
    size_t num_bytes        - (in) Size of memory pointed to by p_inst, in bytes.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int zcs_init(void *p_inst, size_t num_bytes)
{
    int i;
    ZCS *p_zcs;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(num_bytes == zcs_query_mem_size());
    
    /* Retrieve the instance pointer */
    p_zcs = (ZCS *) p_inst;

    /* Initialize the crossing count index */
    p_zcs->crossing_count_idx = 0;

    /* Initialize the crossing counts */
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        p_zcs->p_crossing_count[i] = 0.0f;
    }
    
    /* Initialize the crossing state */
    p_zcs->crossing_state = POSITIVE;

    return 0;
}

/*******************************************************************************************
int zcs_process_block(void *p_inst, float *p_audio, int num_samples)
-Purpose:
    Calculate the zero crossing skew feature for a block of audio.
-Parameters:
    void *p_inst        - (in/out) Pointer to an instance.
    float *p_audio      - (in) Pointer to a block of audio.
    int num_samples     - (in) Number of samples contained in p_audio.  Required to be ZCS_BLOCK_SIZE.
-Returns:
    Returns 0 on success, or -1 on failure.
********************************************************************************************/
int zcs_process_block(void *p_inst, float *p_audio, int num_samples)
{
    int     i;
    int     cross_counter;
    int     *p_crossing_state;
    ZCS     *p_zcs;

    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    ERR_CHECKARG(p_audio != NULL);
    ERR_CHECKARG(num_samples == ZCS_BLOCK_SIZE);
    
    /* Retrieve the instance pointer */
    p_zcs = (ZCS *) p_inst;

    /* Get extra payload info */
    p_crossing_state = &(p_zcs->crossing_state);

    /* Calculate the zero crossing rate */
    cross_counter = 0;
    for (i = 0; i < ZCS_BLOCK_SIZE; i++)
    {
        if ((*p_crossing_state == NEGATIVE) &&
            (p_audio[i] > 0.0f)
            )
        {
            /* Positive going signal */
            *p_crossing_state = POSITIVE;
            cross_counter++;
        }
        else if ((*p_crossing_state == POSITIVE) &&
                 (p_audio[i] < 0.0f)
                 )
        {
            /* Negative going signal */
            *p_crossing_state = NEGATIVE;
            cross_counter++;
        }
    }

    /* Save the crossing count */
    p_zcs->p_crossing_count[p_zcs->crossing_count_idx] = (float) cross_counter;
    p_zcs->crossing_count_idx++;
    p_zcs->crossing_count_idx &= (BLOCKS_PER_FRAME - 1);

    return 0;
}

/*******************************************************************************************
float zcs_process_feature_frame(void *p_inst)
-Purpose:
    Calculate the zero crossing skew feature for a frame.
-Parameters:
    void *p_inst  - (in/out) Pointer to an instance.
-Returns:
    Returns the feature, or 0.0 on error
********************************************************************************************/
float zcs_process_feature_frame(void *p_inst)
{
    int     i;
    float   return_value;
    float   mean;
    float   variance;
    float   skew;
    float   sign;
    float   temp;
    float   temp_skew;
    float   std_dev;
    float   norm;
    ZCS     *p_zcs;
    
    /* Check arguments */
    ERR_CHECKARG(p_inst != NULL);
    
    /* Retrieve the instance pointer */
    p_zcs = (ZCS *) p_inst;

    /* Calculate the mean zero crossing rate */
    mean = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        mean += (p_zcs->p_crossing_count[i]);
    }
    mean /= (float) (BLOCKS_PER_FRAME);

    /* Calculate the variance and skew of the crossing rate */
    variance = 0.0f;
    skew = 0.0f;
    for (i = 0; i < BLOCKS_PER_FRAME; i++)
    {
        temp = p_zcs->p_crossing_count[i] - mean;
        temp_skew = temp;
        temp *= temp;
        temp_skew *= temp;

        variance += temp;
        skew += temp_skew;
    }
    variance /= (float) (BLOCKS_PER_FRAME);
    skew /= (float) (BLOCKS_PER_FRAME);

    /* Calculate the zero crossing skew feature */
    sign = (skew >= 0) ? 1.0f : -1.0f;
    skew = (float) (fabs(skew));

    if (skew > 0.0f && variance > 0.0f)
    {
        std_dev = (float) (sqrt(variance));
        norm = variance * std_dev;

        return_value = sign * skew / norm;

    }
    else
    {
        return_value = 0.0f;
    }

    return return_value;
}
